// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "attr_enum.h"
#include "group_enum.h"
#include "TLibDelay"
#include "LLoadFuncParse.h"
#include "LibLoadMsg.h"
#include "NCell"
#include "TGenOpCnd"
#include <stdlib.h>

#include <cstdlib>
#ifdef RDIPF_lnx
#include <unistd.h>
#endif

static TLibDelaySense
getDelaySense(UKeyword k)
{
  switch (k) {
  case K_non_unate : return TLDS_non_unate;
  case K_positive_unate : return TLDS_positive_unate;
  case K_negative_unate : return TLDS_negative_unate;
  default : return TLDS_non_unate;
  }
}

static TLibArcType
getArcType(UKeyword k)
{
  switch (k) {
  case K_combinational : return TLAT_combinational;
  case K_combinational_rise : return TLAT_combinational_rise;
  case K_combinational_fall : return TLAT_combinational_fall;
  case K_three_state_disable : return TLAT_three_state_disable;
  case K_three_state_disable_rise : return TLAT_three_state_disable_rise;
  case K_three_state_disable_fall : return TLAT_three_state_disable_fall;
  case K_three_state_enable : return TLAT_three_state_enable;
  case K_three_state_enable_rise : return TLAT_three_state_enable_rise;
  case K_three_state_enable_fall : return TLAT_three_state_enable_fall;
  case K_rising_edge : return TLAT_rising_edge;
  case K_falling_edge : return TLAT_falling_edge;
  case K_preset : return TLAT_preset;
  case K_clear : return TLAT_clear;
  case K_hold_rising : return TLAT_hold_rising;
  case K_hold_falling : return TLAT_hold_falling;
  case K_setup_rising : return TLAT_setup_rising;
  case K_setup_falling : return TLAT_setup_falling;
  case K_recovery_rising : return TLAT_recovery_rising;
  case K_recovery_falling : return TLAT_recovery_falling;
  case K_skew_rising : return TLAT_skew_rising;
  case K_skew_falling : return TLAT_skew_falling;
  case K_removal_rising : return TLAT_removal_rising;
  case K_removal_falling : return TLAT_removal_falling;
  case K_min_pulse_width : return TLAT_min_pulse_width;
  case K_minimum_period : return TLAT_minimum_period;
  case K_max_clock_tree_path : return TLAT_max_clock_tree_path;
  case K_min_clock_tree_path : return TLAT_min_clock_tree_path;
  case K_non_seq_setup_rising : return TLAT_non_seq_setup_rising;
  case K_non_seq_setup_falling : return TLAT_non_seq_setup_falling;
  case K_non_seq_hold_rising : return TLAT_non_seq_hold_rising;
  case K_non_seq_hold_falling : return TLAT_non_seq_hold_falling;
  case K_nochange_high_high : return TLAT_nochange_high_high;
  case K_nochange_high_low : return TLAT_nochange_high_low;
  case K_nochange_low_high : return TLAT_nochange_low_high;
  case K_nochange_low_low : return TLAT_nochange_low_low;
  default :     
    TLIBMSG(ignoringArcType, k);
    return TLAT_unknown;
  }
}

void
LLoad::processTimingAttr(LCell& lcell, LibGroup* ltiming, TLibDelay* d, int outCount,
			 LLoadPinMap& pinMap, LLoadBundleMap* bundles,
			 LLoadBusMap* busses, UAggType* busType,
			 int grpIdx, const char*& relatedPin,
			 const char*& relatedOutPin)
{
  LibAttrIter aiter;
  LibAttr* attr;
  TLib& tlib = *d_lib.tlib();
  TLibDelayType ftype = TLDT_none, rtype = TLDT_none;
  TLibDelayLinSlope rise;
  TLibDelayLinSlope fall;
  bool hasDefault = false;
  UTrace traceDefault;
  UTrace traceCond;
  Cudd& bddMgr = d_lib.bddMgr();

  aiter = ltiming->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_default_timing                     : // boolean;
      hasDefault = attr->value()->bool_val();
      if (hasDefault) {
	traceDefault.add(attr->filename(), attr->lineno());
      }
      break;
    case LIBERTY_ATTRENUM_fall_resistance                    : // float;
      if (ftype < TLDT_linear) {
	ftype = TLDT_linear;
      }
      fall.resistance(tlib.convertRes(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_intrinsic_fall                     : // float;
      if (ftype < TLDT_intrinsic) {
	ftype = TLDT_intrinsic;
      }
      fall.intrinsic(tlib.convertTime(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_intrinsic_rise                     : // float;
      if (rtype < TLDT_intrinsic) {
	rtype = TLDT_intrinsic;
      }
      rise.intrinsic(tlib.convertTime(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_mode                               : // ( string, string );
      { 
	UTrace t(attr->filename(), attr->lineno());
	traceCond.add(t);
	if (d->condition().ptr()) {
	  LIBLOADMSGT(multipleConditions, traceCond, lcell.ncell().name().str());
	} else {  
	  LibValIter vIter = attr->valueIter();
	  UString modeName = vIter++.object()->string_val();
	  const char* valStr  = vIter.object()->string_val();
	  char* str = new char[strlen(valStr) + 1];
	  char* val;
	  strcpy(str, valStr);
	  BDD cond = bddMgr.bddZero();
	  for (val = strtok(str, ", \\\t\n"); val; val = strtok(NULL, ", \\\t\n")) {
	    BDD cond1 = lcell.findModeValue(modeName, val, t);
	    if (cond1.ptr()) {
	      cond |= cond1;
	    }
	  }
	  delete [] str;
	  uprintf ("setting timing arc condition to mode %s value %s\n", modeName.str(), valStr);
	  d->setDefault(false);
	  d->condition(cond);
	}
      }
      break;
    case LIBERTY_ATTRENUM_related_pin                        : // virtual_attribute;
      relatedPin = attr->value()->string_val();
      break;
    case LIBERTY_ATTRENUM_related_output_pin                 : // virtual_attribute;
      relatedOutPin = attr->value()->string_val();
      break;
    case LIBERTY_ATTRENUM_rise_resistance                    : // float;
      if (rtype < TLDT_linear) {
	rtype = TLDT_linear;
      }
      rise.resistance(tlib.convertRes(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_slope_fall                         : // float;
      ftype = TLDT_linSlope;
      fall.slope((float)(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_slope_rise                         : // float;
      rtype = TLDT_linSlope;
      rise.slope((float)(attr->value()->double_val()));
      break;
    case LIBERTY_ATTRENUM_timing_sense                       : // enum( non_unate, negative_unate, or positive_unate );
      d->sense(getDelaySense
		     ((UKeyword)UString(attr->value()->string_val()).keyword()));
      break;
    case LIBERTY_ATTRENUM_timing_type                        : // enum( three_state_disable, retaining_time, non_seq_setup_falling, skew_rising, nochange_low_low, nochange_high_low, skew_falling, combinational_fall, nochange_low_high, nochange_high_high, combinational_rise, rising_edge, three_state_enable, preset, setup_rising, three_state_disable_fall, hold_falling, non_seq_hold_rising, minimum_period, removal_falling, non_seq_setup_rising, three_state_enable_fall, clear, non_seq_hold_falling, three_state_disable_rise, recovery_rising, three_state_enable_rise, setup_falling, falling_edge, min_pulse_width, combinational, recovery_falling, removal_rising, or hold_rising );
      d->type(getArcType
		    ((UKeyword)UString(attr->value()->string_val()).keyword()));
      break;
    case LIBERTY_ATTRENUM_when                               : // virtual_attribute;
      // set default to false for now; if there is a separate default statement,
      // this will be taken care of later
      traceCond.add(attr->filename(), attr->lineno());
      if (d->condition().ptr()) {
	LIBLOADMSGT(multipleConditions, traceCond, lcell.ncell().name().str());
      } else {
	d->setDefault(false);
	d->condition(parseFunc(attr->value()->string_val(), bddMgr, pinMap, bundles,
			       busses, busType, grpIdx, outCount));
      }
      break;
    case LIBERTY_ATTRENUM_edge_rate_sensitivity_f0           : // float;
    case LIBERTY_ATTRENUM_edge_rate_sensitivity_f1           : // float;
    case LIBERTY_ATTRENUM_edge_rate_sensitivity_r0           : // float;
    case LIBERTY_ATTRENUM_edge_rate_sensitivity_r1           : // float;
    case LIBERTY_ATTRENUM_hold_coefficient                   : // float;
    case LIBERTY_ATTRENUM_poly_convert                       : // boolean;
    case LIBERTY_ATTRENUM_related_bus_equivalent             : // virtual_attribute;
    case LIBERTY_ATTRENUM_related_bus_pins                   : // virtual_attribute;
    case LIBERTY_ATTRENUM_sdf_cond                           : // string;
    case LIBERTY_ATTRENUM_sdf_cond_end                       : // string;
    case LIBERTY_ATTRENUM_sdf_cond_start                     : // string;
    case LIBERTY_ATTRENUM_sdf_edges                          : // enum( end_edge, start_edge, noedge, or both_edges );
    case LIBERTY_ATTRENUM_setup_coefficient                  : // float;
    case LIBERTY_ATTRENUM_steady_state_resistance_above_high : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_below_low  : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_float_max  : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_float_min  : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_high       : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_high_max   : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_high_min   : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_low        : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_low_max    : // float( > 0.0 );
    case LIBERTY_ATTRENUM_steady_state_resistance_low_min    : // float( > 0.0 );
    case LIBERTY_ATTRENUM_when_end                           : // virtual_attribute;
    case LIBERTY_ATTRENUM_when_start                         : // virtual_attribute;
    case LIBERTY_ATTRENUM_fall_delay_intercept               : // ( integer, list );
    case LIBERTY_ATTRENUM_fall_pin_resistance                : // ( integer, list );
    case LIBERTY_ATTRENUM_rise_delay_intercept               : // ( integer, list );
    case LIBERTY_ATTRENUM_rise_pin_resistance                : // ( integer, list );
    default                                                  :
      if (warnNoSupport()) {
	UTrace loc(attr->filename(), attr->lineno());
	// "%s '%s' (in %s group) not supported - ignored"
	LIBLOADMSGT(notSupported, loc, "Attribute", attr->name().str(), "timing");
      }
      break;
    }
  }
  // d->default is initialized to true at construction;
  // if timing has a when condition; d->default will be reset to false
  // in the above switch statement (see case for LIBERTY_ATTRENUM_when)
  // however, if timing also has a default attribute set to true, then
  // d->default should be set to true
  if (hasDefault) {
    d->setDefault(true);
    if (d->condition().ptr()) {
      LIBLOADMSGT(defaultArcWithCondition, traceDefault);
    }
  }
  if (rtype > TLDT_none) {
    uassert(d->delayType(TEdge::rise) == TLDT_none);
    d->delay(TEdge::rise, &rise, rtype);
  }
  if (ftype > TLDT_none) {
    uassert(d->delayType(TEdge::fall) == TLDT_none);
    d->delay(TEdge::fall, &fall, ftype);
  }
}

void
LLoad::processValueList(LibValEList* valList, TLibDelayTable* ttable)
{
  LibVal* v;
  TLib& tlib = *d_lib.tlib();
  TValueArray* valArr;
  int count[3] = {1, 1, 1};
  int index[3] = {0, 0, 0};
  int pos[3];
  int i, idx;
  TLutIndexTable* idxTable = ttable->idxTable();
  LibValIter viter = valList->begin();
  char* val;
  char* str = NULL;
  int len = 0, l;
  const char* valStr;

  if (idxTable) {
    for (i = idxTable->varCount() - 1; i >= 0; i--) {
      pos[i] = idxTable->varPos(i);
    }
    if (idxTable->varCount() == 3) {
      count[2] = idxTable->indices(2).count();
    }
    if (idxTable->varCount() > 1) {
      count[1] = idxTable->indices(1).count();
    }
    count[0] = idxTable->indices(0).count();
  }
  // need to dynamically allocate valArr so that the array will be
  // allocated in the tlib pool, since this array will eventually
  // be given to some tlib pin class
  valArr = new IN_SAME_POOL_AS(&tlib) TValueArray(count[0]*count[1]*count[2]);
  while ((v = viter++)) {
    valStr = v->string_val();
    l = strlen(valStr) + 1;
    if (l > len) {
      delete [] str;
      len = l;
      str = new char[len];
    }
    strcpy(str, valStr);
    for (val = strtok(str, ", \\\t\n"); val; val = strtok(NULL, ", \\\t\n")) {
      idx = (((index[0] * count[1]) + index[1]) * count[2]) + index[2]; 
      (*valArr)[idx] = tlib.convertTime(strtod(val, NULL));
      if (!idxTable) break;
      for (i = idxTable->varCount() - 1; i >= 0; i--) {
	index[pos[i]]++;
	if (index[pos[i]] == count[pos[i]]) {
	  index[pos[i]] = 0;
	} else {
	  break;
	}
      }
    }
  }
  uassert (!index[0] && !index[1] && !index[2]);
  ttable->grab(valArr);
  delete valArr;
  delete [] str;
}

void
LLoad::processTimingTable(LibGroup* ltable, TLibDelayTable* ttable)
{
  LibErrorT err;
  TLib& tlib = *d_lib.tlib();
  const UString& tableName = ltable->firstName();
  int i;
  LibAttr* attr;
  LibValEList* valStrList;

  if (tableName.keyword() != K_scalar) {
    TLutTemplate* temp = tlib.findLutTemplate(tableName);
    UKeyword index[3] = {K_index_1, K_index_2, K_index_3};
    const char* idxStr[3] = {NULL, NULL, NULL};

    for (i = temp->varCount() - 1; i >= 0; i--) {
      if ((attr = ltable->findAttr(index[i], err))) {
	idxStr[i] = attr->firstValue()->string_val();
      }
    }
    ttable->idxTable(temp->findOrAddIndexTable(idxStr));
  }
  attr = ltable->findAttr(K_values, err);
  valStrList = attr->valueList();
  processValueList (valStrList, ttable);
}

void
LLoad::processTimingSubgroup(LibGroup* ltiming, TLibDelay* d)
{
  LibGroupIter giter;
  LibGroup* subgroup;
  TLibDelayType rtype = TLDT_none, ftype = TLDT_none;
  TLibDelayTablePair rise, fall;
  int rtrans = 0, ftrans = 0;

  giter = ltiming->groupIter();
  while ((subgroup = giter++)) {
#ifdef RT_FPGA
    exit(1);
#endif

    switch (subgroup->typeID()) {
    case LIBERTY_GROUPENUM_cell_fall                                   : // ( cell_fall_name )
      uassert(ftype == TLDT_none);
      ftype = TLDT_cell;
      processTimingTable(subgroup, fall.table());
      break;
    case LIBERTY_GROUPENUM_cell_rise                                   : // ( cell_rise_name )
      uassert(rtype == TLDT_none);
      rtype = TLDT_cell;
      processTimingTable(subgroup, rise.table());
      break;      
    case LIBERTY_GROUPENUM_fall_constraint                             : // ( fall_constraint_name )
      uassert(ftype == TLDT_none);
      ftype = TLDT_constraint;
      processTimingTable(subgroup, fall.table());
      break;
    case LIBERTY_GROUPENUM_fall_propagation                            : // ( fall_propagation_name )
      uassert(ftype == TLDT_none);
      ftype = TLDT_propagation;
      processTimingTable(subgroup, fall.table());
      break;
    case LIBERTY_GROUPENUM_fall_transition                             : // ( fall_transition_name )
      uassert(!ftrans);
      ftrans = 1;
      processTimingTable(subgroup, fall.transTable());
      break;
    case LIBERTY_GROUPENUM_rise_constraint                             : // ( rise_constraint_name )
      uassert(rtype == TLDT_none);
      rtype = TLDT_constraint;
      processTimingTable(subgroup, rise.table());
      break;
    case LIBERTY_GROUPENUM_rise_propagation                            : // ( rise_propagation_name )
      uassert(rtype == TLDT_none);
      rtype = TLDT_propagation;
      processTimingTable(subgroup, rise.table());
      break;
    case LIBERTY_GROUPENUM_rise_transition                             : // ( rise_transition_name )
      uassert(!rtrans);
      rtrans = 1;
      processTimingTable(subgroup, rise.transTable());
      break;
    case LIBERTY_GROUPENUM_cell_degradation                            : // ( cell_degradation_name )
    case LIBERTY_GROUPENUM_noise_immunity_above_high                   : // ( noise_immunity_above_high_name )
    case LIBERTY_GROUPENUM_noise_immunity_below_low                    : // ( noise_immunity_below_low_name )
    case LIBERTY_GROUPENUM_noise_immunity_high                         : // ( noise_immunity_high_name )
    case LIBERTY_GROUPENUM_noise_immunity_low                          : // ( noise_immunity_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_height_above_high          : // ( propagated_noise_height_above_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_height_below_low           : // ( propagated_noise_height_below_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_height_high                : // ( propagated_noise_height_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_height_low                 : // ( propagated_noise_height_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_peak_time_ratio_above_high : // ( propagated_noise_peak_time_ratio_above_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_peak_time_ratio_below_low  : // ( propagated_noise_peak_time_ratio_below_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_peak_time_ratio_high       : // ( propagated_noise_peak_time_ratio_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_peak_time_ratio_low        : // ( propagated_noise_peak_time_ratio_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_width_above_high           : // ( propagated_noise_width_above_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_width_below_low            : // ( propagated_noise_width_below_low_name )
    case LIBERTY_GROUPENUM_propagated_noise_width_high                 : // ( propagated_noise_width_high_name )
    case LIBERTY_GROUPENUM_propagated_noise_width_low                  : // ( propagated_noise_width_low_name )
    case LIBERTY_GROUPENUM_retain_fall_slew                            : // ( retain_fall_slew_name )
    case LIBERTY_GROUPENUM_retain_rise_slew                            : // ( retain_rise_slew_name )
    case LIBERTY_GROUPENUM_retaining_fall                              : // ( retaining_fall_name )
    case LIBERTY_GROUPENUM_retaining_rise                              : // ( retaining_rise_name )
    case LIBERTY_GROUPENUM_steady_state_current_high                   : // ( steady_state_current_high_name )
    case LIBERTY_GROUPENUM_steady_state_current_low                    : // ( steady_state_current_low_name )
    case LIBERTY_GROUPENUM_steady_state_current_tristate               : // ( steady_state_current_tristate_name )
    default:
      if (warnNoSupport()) {
	UTrace loc(subgroup->filename(), subgroup->lineno());
	// "%s '%s' (in %s group) not supported - ignored"
	LIBLOADMSGT(notSupported, loc, "Group", subgroup->type().str(), "timing");
      }
      break;
    }
  }
  if (rtrans) {
    uassert(rtype == TLDT_cell || rtype == TLDT_propagation);
    if (rtype == TLDT_cell) {
      rtype = TLDT_cellNTrans;
    } else {
      rtype = TLDT_propNTrans;
    }
  }
  if (ftrans) {
    uassert(ftype == TLDT_cell || ftype == TLDT_propagation);
    if (ftype == TLDT_cell) {
      ftype = TLDT_cellNTrans;
    } else {
      ftype = TLDT_propNTrans;
    }
  }
  if (rtype > TLDT_none) {
    uassert(d->delayType(TEdge::rise) == TLDT_none);
    d->delay(TEdge::rise, &rise, rtype);
  }
  if (ftype > TLDT_none) {
    uassert(d->delayType(TEdge::fall) == TLDT_none);
    d->delay(TEdge::fall, &fall, ftype);
  }
}

static const char* 
arcTypeString(TArcType t)
{
  switch (t) {
  case TAT_comb   : return "combinational";
  case TAT_trigger: return "trigger";
  case TAT_setup  : return "setup";
  case TAT_check  : return "check";
  case TAT_none   : return "none";
  }
  return "?";
}

static void
updateDelaySense(TLibDelay* d, BDD f, int inputIdx)
{
  if (!f.ptr() || d->sense() != TLDS_non_unate || d->type() != TLAT_combinational) {
    return;
  }
  const BDD& cond = d->condition();
  if (cond.ptr()) {
    f = f.Restrict(cond);
  }
  if (f.Increasing(inputIdx).IsOne()) {
    d->sense(TLDS_positive_unate);
  } else if (f.Decreasing(inputIdx).IsOne()) {
    d->sense(TLDS_negative_unate);
  }
}

void
LLoad::processRelatedPin(TLibDelay* orig, const char* relatedPinStr,
			 TLibTimingPinC* tpin, int toIdx, LCell& lcell,
			 LLoadPinMap& pinMap, LLoadBundleMap* bundles,
			 LLoadBusMap* busses, UAggType* busType,
			 int grpIdx, uint_t& delayCount,
			 bool dontAddIfExist)
{
  NCell cell = lcell.ncell();
  char* str;
  char* rPinName;
  TLibDelay* d;
  int frIdx;
  TLibPin toPin = lcell.pin(toIdx);
  BDD pinFunc;
  UStrBuf buf;

  if (toPin.isOutput()) {
    pinFunc = toPin.function();
  }
  str = new char[strlen(relatedPinStr) + 1];
  buf = UNaming::descapeForLiberty(relatedPinStr);
  strcpy(str, buf.str());
  for (rPinName = strtok(str, " \\\t\n"); rPinName; rPinName = strtok(NULL, " \\\t\n")) {
    frIdx = pinMap[rPinName] + LLoadIndexAdjust(rPinName, bundles, busses, busType, grpIdx);
    if (frIdx == toIdx) {
      // cannot handle self-arc
      continue;
    }
    d = new IN_SAME_POOL_AS(orig) TLibDelay(*orig);
    updateDelaySense(d, pinFunc, frIdx - lcell.outPinCount()); 
    int rc = tpin->add(d, frIdx, toIdx, d_libCorner, cell, delayCount, dontAddIfExist);
    if (rc == 1) {
      TLIBMSG(inconsistentArcType, (str_t)cell.name(),
	      (str_t)cell.pinName(frIdx), (str_t)cell.pinName(toIdx),
	      string(d->type()),
	      arcTypeString(tpin->getTimingType(frIdx)));
      delete d;
    } else if (rc == 2) {
      TLIBMSG(pinTimingOverwrites, (str_t)cell.name(),
	      (str_t)cell.pinName(frIdx), (str_t)cell.pinName(toIdx));
      delete d;
    } else {
      d->grabDelayOwnershipFrom(orig);
    }
  }
  delete [] str;
}

void
LLoad::processTimingGroup(LibGroup* ltiming, TLibTimingPinC* tpin, int tIdx,
			  int outCount, LCell& lcell, LLoadPinMap& pinMap, 
			  LLoadBundleMap* bundles, LLoadBusMap* busses, UAggType* busType,
			  int grpIdx, uint_t& delayCount, bool dontAddIfExist)
{
  NCell cell = lcell.ncell();
  TLibDelay* d = new IN_SAME_POOL_AS(d_lib.tlib()) TLibDelay;
  const char* relatedPin = NULL;
  const char* relatedOutPin = NULL;

  processTimingAttr(lcell, ltiming, d, outCount, pinMap, bundles, busses, busType,
		    grpIdx, relatedPin, relatedOutPin);
  if (relatedOutPin) {
    int idx = pinMap[relatedOutPin] +
              LLoadIndexAdjust(relatedOutPin, bundles, busses, busType, grpIdx);
    d->relatedOut(idx);
  }
  if (relatedPin) {
    processTimingSubgroup(ltiming, d);
    processRelatedPin(d, relatedPin, tpin, tIdx, lcell, pinMap, bundles,
		      busses, busType, grpIdx, delayCount, dontAddIfExist);
  }
  delete d;
}

void
LLoad::processPinSubgroup(const LibGroup* lpin, int tidx,
			  LCell& lcell, LLoadPinMap& pinMap, LLoadBundleMap* bundles,
			  LLoadBusMap* busses, UAggType* busType,
			  int grpIdx, uint_t& delayCount, bool dontAddIfExist)
{
  LibGroupIter giter;
  LibGroup* subgroup;

  TLibTimingPinC* tpin = (TLibTimingPinC*)lcell.pin(tidx).ptr();
  int outCount = lcell.outPinCount();
  giter = lpin->groupIter();
  while ((subgroup = giter++)) {
    switch (subgroup->typeID()) {
    case LIBERTY_GROUPENUM_timing                      : // ( [timing_name] )
      processTimingGroup (subgroup, tpin, tidx, outCount, lcell, pinMap,
			  bundles, busses, busType, grpIdx, delayCount, dontAddIfExist);
      break;
    case LIBERTY_GROUPENUM_electromigration            : // ( [electromigration_name] )
    case LIBERTY_GROUPENUM_hyperbolic_noise_above_high : // ( [hyperbolic_noise_above_high_name] )
    case LIBERTY_GROUPENUM_hyperbolic_noise_below_low  : // ( [hyperbolic_noise_below_low_name] )
    case LIBERTY_GROUPENUM_hyperbolic_noise_high       : // ( [hyperbolic_noise_high_name] )
    case LIBERTY_GROUPENUM_hyperbolic_noise_low        : // ( [hyperbolic_noise_low_name] )
    case LIBERTY_GROUPENUM_internal_power              : // ( [internal_power_name] )
    case LIBERTY_GROUPENUM_memory_read                 : // ( [memory_read_name] )
    case LIBERTY_GROUPENUM_memory_write                : // ( [memory_write_name] )
    case LIBERTY_GROUPENUM_min_pulse_width             : // ( [min_pulse_width_name] )
    case LIBERTY_GROUPENUM_minimum_period              : // ( [minimum_period_name] )
    case LIBERTY_GROUPENUM_tlatch                      : // ( tlatch_name )
    default:
      if (warnNoSupport()) {
	UTrace loc(subgroup->filename(), subgroup->lineno());
	// "%s '%s' (in %s group) not supported - ignored"
	LIBLOADMSGT(notSupported, loc, "Group", subgroup->type().str(), "pin");
      }
      break;
    }
  }
  processDefaultArcs(tpin, lcell.totalPinCount());
}

/// Compute the condition for the default arc(s) (if any)
void
LLoad::processDefaultArcs(TLibTimingPinC* tpin, uint_t pinCount)
{
  Cudd* mgr = &d_lib.bddMgr();
  for (uint_t j = 0; j < pinCount; j++) {
    if (TLibDelayList* list = tpin->delays(j, MVCRNR_INITIAL, d_libCorner)) {
      TLibDelayIter delayIter = list->begin();
      bool hasDefault = false;
      // skip default arcs (if any)
      for (; delayIter; delayIter++) {
	TLibDelay* lDelay = delayIter.object();
	if (lDelay->isDefault()) {
	  hasDefault = true;
	} else {
	  break;
	}
      }
      if (hasDefault) {
	// compute the condition for the default
	BDD bdd = mgr->bddOne();
	for (; delayIter; delayIter++) {
	  TLibDelay* lDelay = delayIter.object();
	  uassert(!lDelay->isDefault());
	  if (lDelay->condition() > bdd) {
	    LIBLOADMSG(arcConditionsOverlap);
	  }
	  bdd &= ~lDelay->condition();
	}
	// set this bdd as the condition to all default arcs
	for (delayIter = list->begin(); delayIter; delayIter++) {
	  TLibDelay* lDelay = delayIter.object();
	  if (lDelay->isDefault()) {
	    if (lDelay->condition().ptr()) {
	      if (lDelay->condition() != bdd) {
		if (lDelay->condition() <= bdd) {
		  LIBLOADMSG(defaultArcConditionExcludes);
		} else {
		  LIBLOADMSG(arcConditionsOverlap);
		}
	      }
	    } else {
	      lDelay->condition(bdd);
	    }
	  } else {
	    break;
	  }
	}
      }
    }
  }  
}
